%skeleton "lalr1-c++17.cc" /* -*- C++ -*- */
%require "3.0.4"
%defines
%define extra_header_prefix {calc-c++17-}  // prefix for the generated headers:
                                           // position.hh, location.hh and stack.hh
%define parser_class_name {calcxx_parser}
%define api.token.constructor
%define api.value.type variant
%define parse.assert

// Optional: use std::deque
// If you remove these two lines, it will use std::vector by default.
// Both work well with this parser.
%define api.stack.include deque
%define api.stack.container {std::deque}

%code requires
{
# include <string>
# include <memory>
class calcxx_driver;
}
// The parsing context.
%param { calcxx_driver& driver }
%locations
%initial-action
{
  // Initialize the initial location.
  @$.begin.filename = @$.end.filename = &driver.file;
};
%define parse.trace
%define parse.error verbose
%code
{
# include "calc-c++17-driver.hh"

// define a printer function for semantic values for std::variant
namespace yy
{

  // yy_print_value: utility function to print semantic values in lexers and parsers
  // (only needed with parse.trace)

// Two versions given here for demonstration, both work the same here.
// You can choose them with this conditional:
#define YY_PRINT_VALUE_SIMPLE
#ifdef YY_PRINT_VALUE_SIMPLE

  // Rather simple version, fully sufficient for this demo.

  template <typename T>
  void yy_print_value (std::ostream& o, const T& i)
  {
    o << i;
  }

  template <typename T>
  void yy_print_value (std::ostream& o, const std::unique_ptr <T>& i)
  {
    if (!i)
      o << "(null)";
    else
      o << *i;
  }

#else

  /*
     This section shows how to declare generic versions of yy_print_value
     for some common use cases. For this demo, this is overkill, but it
     can serve as a basis for more complicated cases:

     - (Smart) pointers will be dereferenced for output, so their contents,
       rather than their addresses, will be shown. If you prefer addresses,
       you can omit this overload. If you prefer something else, such as
       outputting "&" before the contents, you can change the pointer
       overload accordingly.

     - Containers will be output element-wise, comma-separated and enclosed
       in "{ ... }". Again, you can change this to your liking.

     - Combinations, such as containers of pointers, pointers to containers,
       containers of containers etc., will use respective combinations of
       these overloads.

     - Types that can be output directly with "<<", will be output this way.
       E.g., std::string, even though it would pass as a container, will be
       output directly, rather than element-wise; same for any user-defined
       output operators for some containers. The only exceptions are (smart)
       pointers, which could also be output directly (as their address), but
       instead their contents are output (see above).

     - If wanted, you can add further generic versions, e.g. for std::pair,
       std::tuple or even std::variant if for some reason you want to use
       nested variants. You should add forward-declarations (see below),
       so combinations work.

     - Of course, you can also add overloads for specific types, e.g. for
       std::string, if you want to output it with special characters quoted
       or something similar. You can either overload yy_print_value, if you
       want this form of printing only for the parser, or overload
       operator<< if you want it in other parts of your code as well.
  */

  // SFINAE structs to detect types that can be output directly with "<<", (smart) pointers, and containers

  template <typename T, typename _ = void> struct has_operator_output: std::false_type { };
  template <typename T> struct has_operator_output <T, std::void_t <decltype (std::cout << std::declval <T> ())>>: std::true_type { };

  template <typename T, typename _ = void> struct has_operator_star: std::false_type { };
  template <typename T> struct has_operator_star <T, std::void_t <decltype (*std::declval <T> ())>>: std::true_type { };

  template <typename T, typename _ = void> struct is_container: std::false_type { };
  template <typename T> struct is_container <T,
    std::void_t <typename T::value_type,
                 typename T::size_type,
                 typename T::iterator,
                 typename T::const_iterator,
                 decltype (std::declval <T> ().size ()),
                 decltype (std::declval <T> ().begin ()),
                 decltype (std::declval <T> ().end ()),
                 decltype (std::declval <T> ().cbegin ()),
                 decltype (std::declval <T> ().cend ())>>: std::true_type { };

  // basic version of yy_print_value, enabled if no overload applies
  template <typename T>
  std::enable_if_t <
    !has_operator_star <T>::value &&
    !(is_container <T>::value && !has_operator_output <T>::value)
  > yy_print_value (std::ostream &o, const T &v)
  {
    o << v;
  }

  // forward-declarations, so combinations work

  template <typename T>
  std::enable_if_t <has_operator_star <T>::value>
  yy_print_value (std::ostream &o, const T &v);

  template <typename T>
  std::enable_if_t <is_container <T>::value && !has_operator_output <T>::value>
  yy_print_value (std::ostream &o, const T &v);

  // implementations

  // pointer version (also for directly printable types)
  template <typename T>
  std::enable_if_t <has_operator_star <T>::value>
  yy_print_value (std::ostream &o, const T &v)
  {
    yy_print_value (o, *v);
  }

  // container version (not for directly printable types)
  template <typename T>
  std::enable_if_t <is_container <T>::value && !has_operator_output <T>::value>
  yy_print_value (std::ostream &o, const T &v)
  {
    o << '{';
    size_t n_ = 0;
    for (auto &i_: v)
      {
        if (n_++)
          o << ", ";
        yy_print_value (o, i_);
      }
    o << '}';
  }

#endif
}
}
%define api.token.prefix {TOK_}
%token
  END  0  "end of file"
  ASSIGN  ":="
  MINUS   "-"
  PLUS    "+"
  STAR    "*"
  SLASH   "/"
  LPAREN  "("
  RPAREN  ")"
;
// We use unique_ptr here only to demonstrate the use of a move-only type
// with the C++17 parser. Plain strings, like in the original calc++ would
// work as well, of course.
%token <std::unique_ptr <std::string>> IDENTIFIER "identifier"
%token <int> NUMBER "number"
%type  <int> exp
%%
%start unit;
unit: assignments exp  { driver.result = $2; };

assignments:
  %empty
| assignments assignment;

assignment:
  "identifier" ":=" exp { driver.variables[*$1] = $3; };

%left "+" "-";
%left "*" "/";
exp:
  exp "+" exp   { $$ = $1 + $3; }
| exp "-" exp   { $$ = $1 - $3; }
| exp "*" exp   { $$ = $1 * $3; }
| exp "/" exp   { $$ = $1 / $3; }
| // pointless mid-rule action just for demo
  "(" exp { $<int>$ = $2; } ")" { $$ = $<int>3; }
| "identifier"  { $$ = driver.variables[*$1]; }
| "number"      // use default action
%%
void
yy::calcxx_parser::error (const location_type& l,
                          const std::string& m)
{
  driver.error (l, m);
}
